require 'uu/os/lang/uniform_hash'
require 'uu/os/application/logger'
require 'uu/os/application/log_helper'

module UU
  module OS
    # Application module
    module Application

      # Class represent application global configuration, configuration is read from configuration files and environment variables.
      # Configuration is merged in this order:
      # * Configuration file <app-root>/config/app.rb
      # * <uu-home>/config/uu-client.properties
      # * Environment variable SERVER_CFG
      # * Configuration made by block {UU::OS::Application::Config#load} defined in file <app-root>/config/app.rb 
      #
      # @example
      #   # static configuration
      #   UU::OS::Application::Config.replace({
      #     :app_credentials => '77-456-2',
      #     :my_param => 'my value'
      #   })
      #   # config params: {:app_credentials => '77-456-2', :my_param => 'my value'}
      #
      #   # get value anywhere in application
      #   UU::OS::Application::Config[:app_credentials] # String '77-456-2'
      #
      #   # merge config
      #   UU::OS::Application::Config.merge!({
      #     :app_credentials => '13-13-2',
      #     :another_param => 13
      #   })
      #   # config params: {:app_credentials => '13-13-2', :my_param => 'my value', :another_param => 13}
      #
      #   # set value
      #   UU::OS::Application::Config[:my_param] # String 'my value'
      #   UU::OS::Application::Config[:my_param] = 'another value'
      #   UU::OS::Application::Config[:my_param] # String 'another value'
      class Config
        extend UU::OS::Application::LogHelper
        # @private
        AUTHN_TOKEN_PARAM = :'uu.os.rest.remote-client-authn-token' # :nodoc:
        private_constant :AUTHN_TOKEN_PARAM

        # Default ttl in seconds. TTL is period of time, while configuration is cached by application.
        DEFAULT_TTL = 3600

        # @private
        @@ctx = UU::OS::Lang::UniformHash.new  # :nodoc:
        # @private
        @@load_callback = nil  # :nodoc:
        # @private
        @@last_update = 0  # :nodoc:
        # @private
        @@ttl = DEFAULT_TTL  # :nodoc:
        # @private
        @@lock = Mutex.new  # :nodoc:
        # @private
        @@reload_disabled = false  # :nodoc:
        # @see Hash#[]
        def self.[](key)
          _reload if reload_necessary?
          @@ctx[key]
        end

        # @see Hash#[]=
        def self.[]=(key, value)
          @@ctx[key] = value
        end

        # @see Hash#merge
        def self.merge!(other_hash, &block)
          @@ctx.merge!(other_hash, &block)
        end

        # @see Hash#replace
        def self.replace(other_hash)
          @@ctx.replace(other_hash)
        end

        # @see Hash#delete
        def self.delete(key, &block)
          @@ctx.delete(key, &block)
        end

        # @see Hash#has_key?
        def self.has_key?(key)
          @@ctx.has_key?(key)
        end

        # @see Hash#clear
        def self.clear
          @@ctx.clear
        end

        # Force immediate configuration reload no matter to ttl expiration. Method waits until configuration is loaded.
        # Reload in done by block defined by {UU::OS::Application::Config#load}.
        #
        # @see UU::OS::Application::Config#load
        # @return [void]
        def self.reload
          _reload(true) 
          return
        end

        # Method takes a block as parameter, this block is used for dynamic configuration loading. Block is called every time
        # when configuration expire (you can set configuration ttl). If configuration expire, this block is called when 
        # application try to read some value (by UU::OS::Application::Config[:param_name]). 
        # 
        # Code used in block for loading must be fast, because if configuration expire so all requests are blocked
        # until configuration is loaded.
        # 
        # If configuration load fails (exception is raised from block), then original configuration is restored and all changes 
        # made on config hash (hash that comes to block) are discarded. The next attempt to reload configuration after failure is made 
        # when normal ttl expire (no special behavior after failure).
        #
        # In block, you must not set values by modifying UU::OS::Application::Config directly (e.g. UU::OS::Application::Config[:param_name] = 'value')
        # but use config hash that comes to the block.
        #
        # There is not guaranty which user (if any) will be logged on session when block starts. Every session change (login or logout) is valid only
        # in this block execution scope and is discarded when block ends.
        #
        # @example
        #   # all necessary requires
        #   require 'uu/os/security/session'
        #   require 'uu/os/property'
        # 
        #   # load configuration from property and merge
        #   # it with original configuration
        #   UU::OS::Application::Config.load do |config|
        #     app_credentials = config[:app_credentials]
        #     UU::OS::Security::Session.login(app_credentials)
        #  
        #     config_property_uri = config[:config_property_uri]
        #     new_config_str = UU::OS::Property.get_value(config_property_uri)
        #     new_config = JSON.parse(new_config_str, :symbolize_names => true)
        #  
        #     config.merge!(new_config)
        #   end
        #
        # @note This method should be defined in <app-root>/config/app.rb file.
        # @yieldparam [UU::OS::Lang::UniformHash] config Hash with actual configuration, you also must use this hash for storing new configuration.
        def self.load(&block)
          return unless block_given?
          @@load_callback = block
        end

        # Returns configuration ttl (time to live) in seconds. Default value is defined by {DEFAULT_TTL} constant.
        # TTL is period of time, while configuration is cached by application.
        #
        # @example
        #   UU::OS::Application::Config.ttl # Fixnum 3600 (means 3600 seconds = 1 hour)
        #
        # @return [Fixnum] ttl
        def self.ttl
          @@ttl
        end

        # Sets configuration ttl (time to live) in seconds. Default value is defined by {DEFAULT_TTL} constant.
        # Value can be set in interval from 1 to {Fixnum} maximum. If you set value to zero or less (e.g. -1) it means infinite ttl and
        # reload is performed only once.
        #
        # @example
        #   # set ttl to 86400 seconds = 1 day 
        #   UU::OS::Application::Config.ttl = 86400 
        #
        # @param [Fixnum] new_ttl New ttl value.
        # @return [void]
        def self.ttl=(new_ttl)
          @@ttl = new_ttl
        end

        private

        # Method invoke load block defined by UU::OS::Application::Config.load.
        # Method is synchronized so only one thread may read configuration in one time.
        #
        # @param [TrueClass, FalseClass] force Flag if reload will be executed even if another thread updates
        #   configuration while this thread is waiting for lock
        #
        # @private
        def self._reload(force = false) # :nodoc:
          # return if no load block was provided or reload is disabled
          return unless @@load_callback
          return if @@reload_disabled
          # we need synchronize this, because more threads may trigger method in same time
          @@lock.synchronize do
            # we check if reload is still necessary, we don't want other threads that waiting for lock (while first thread performs reload)
            # to perform load block again after first thread release lock
            return if (!force) && (!reload_necessary?)
            # save actual session tree, if they use session.login in load block 
            actual_session = Thread.current[AUTHN_TOKEN_PARAM]
            begin
              # this is protection to endless loop, if Config[:param] is called from block and ttl is expired
              # so this method is called again and endless loop occurs
              @@reload_disabled = true
              # shallow copy of configuration because if block fails so configuration is not in inconsistent state
              # (half new and half old)
              temp_config = @@ctx.clone
              # call reload block
              @@load_callback.call(temp_config)
              # set new config
              @@ctx = temp_config
            rescue Exception => ex
              msg = "Error while executing configuration block. This block is defined by"\
              " UU::OS::Application::Config.load method (typically in <app-root>/config/app.rb file)."
              logger = UU::OS::Application::Logger.new(self.class)
              log_error(logger,ex, msg)
            ensure
              @@reload_disabled = false
              # set last update time
              @@last_update = Time.now.to_i
              # restore session tree to state before block is yielded
              # so we don't care if some session.login was called in block
              Thread.current[AUTHN_TOKEN_PARAM] = actual_session
            end
          end
        end

        # @private
        def self.reload_necessary? # :nodoc:
          # check if configuration expired
          return (@@ttl > 0) && (!@@load_callback.nil?) && ((@@last_update + @@ttl) < Time.now.to_i)
        end

        # @private
        # def self.log_error(exception = nil, msg = nil) # :nodoc:
        #   message, exception = exception, message if message.kind_of?(String)
        #   # this project has no direct dependency to uu_os_framework (where UU::OS::Logger is)
        #   # so we try if UU::OS::Logger is available and log where
        #   # TODO this if condition do not work fine, because sometimes UU::OS.const_defined?(:Logger) returns true
        #   # but UU::OS::Logger doesn't exists.
        #   if (UU::OS.const_defined?(:Logger))
        #     log_error_uu_log(exception, msg)
        #   else
        #     log_error_std_err(exception, msg)
        #   end
        # end

        # @private
        # def self.log_error_std_err(exception = nil, message = nil) # :nodoc:
        #   msg = "ERROR #{Time.now} "
        #   if !message.nil? && !message.empty?
        #     msg << message.to_s
        #   end
        #   msg << "\n  " if !exception.nil?
        #   if !exception.nil?
        #     msg << "#{exception.class}"
        #     msg <<  (exception.respond_to?(:id) ? "[#{exception.id}]:" : " : ")
        #     msg <<  "#{exception.message}"
        #     trace = exception.backtrace
        #     msg << "\n  " << trace.join("\n  ") if !trace.nil? && !trace.empty?
        #   end
        #   STDERR.puts(msg)
        # end

        # Logs using {UU::OS::Logger}.
        # @private
        # def self.log_error_uu_log(exception = nil, message = nil) # :nodoc:
        #   begin
        #     log = UU::OS::Logger.new(self.class)
        #     log.error(message, exception)
        #   rescue Exception => ex
        #     # TODO temporary fix
        #     # this is needed because sometimes UU::OS.const_defined?(:Logger) returns true
        #     # but UU::OS::Logger doesn't exists.
        #     log_error_std_err(exception, message)
        #   end
        # end

        # instance is not allowed
        private_class_method :new
      end
    end
  end
end