require 'stringio'
require 'uu/os/server/abstract_middleware'
require 'uu/os/server/request'
require 'uu/os/server/immediate_response'
require 'uu/os/server/json_parser'
require 'uu/os/server/query_parser'
require 'uu/os/server/multipart_parser'

module UU
  module OS
    module Application
      class ParameterParseMiddleware < UU::OS::Server::AbstractMiddleware

        def initialize(app)
          super(app)
          @query_parser = default_query_parser
          @body_parsers = {
              "application/json" => default_json_parser,
              "multipart/form-data" => default_multipart_parser
          }
        end

        def call(env)
          request = UU::OS::Server::Request.new(env)
          parse_query(request)
          parse_body(request)
          @app.call(env)
        end

        private

        # @private
        def parse_query(request)
          parser = @query_parser
          return unless parser

          begin
            query_params = parser.call(request)
          rescue => e
            raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::INVALID_REQUEST,
                                                            "Invalid request: #{e.message}")
          end

          return unless query_params
          request.env['rack.request.query_hash'] = query_params
          request.env['rack.request.query_string'] = request.env['QUERY_STRING'].to_s
        end

        # @private
        def parse_body(request)
          return if request.get?
          media_type = request.media_type

          parser = @body_parsers.detect { |content_type, parser| media_type.match(content_type) } if media_type

          if parser.nil?
            supported_types = @body_parsers.keys.join(", ")
            # TODO we need exception with status code ?
            #raise UU::OS::Server::ImmediateResponse.new(415, {}, "Unsupported content type, expected types: #{supported_types}")
            raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::INVALID_REQUEST,
                                                            "Invalid content type, expected one from: #{supported_types}.")
          end
          begin
            body_params = parser.last.call(request)
          rescue => e
            raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::INVALID_REQUEST,
                                                            "Invalid request: #{e.message}")
          end
          return unless body_params

          request.env["rack.request.form_hash"] = body_params
          request.env["rack.request.form_input"] = request.env["rack.input"]
        end

        # @private
        def default_query_parser
          return proc do |request|
            params = UU::OS::Server::QueryParser.parse(request.query_string)
            return {} if params.nil?
            wrapped = params.delete(parameters_parameter_name)
            if !wrapped.nil? && !wrapped.kind_of?(Hash)
              raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::INVALID_REQUEST,
                                                              "Invalid request: parameter #{parameters_parameter_name} must contain key-value structure")
            else
              params.merge!(wrapped) if !wrapped.nil?
            end
            params
          end
        end

        # @private
        def default_json_parser
          return proc do |request|
            params = UU::OS::Server::JsonParser.parse(request.body)
            if !params.kind_of?(Hash)
              raise TypeError, 'JSON does not represent key-value structure'
            end
            params
          end
        end

        # @private
        def default_multipart_parser
          return proc do |request|
            binary_name = binary_name_parameter_name if binary_expected?(request)
            params = UU::OS::Server::MultipartParser.parse(request, binary_name: binary_name)

            # this is ugly hack to stop frameworks to read input for streaming
            request.env["rack.input"] = StringIO.new("")
            request.env["CONTENT_LENGTH"] = "0"
            params
          end
        end

        # @private
        def binary_expected?(request)
          parts = request.url.split('/')
          return request.post? && (parts[-1] =~ /.*Data$/ || parts[-2] =~ /.*Data$/)
        end

        # @private
        def binary_name_parameter_name
          return :data
        end

        # Returns name of parameter which is used for wrapping of parameters
        # sent in get requests (defaults to +:uuDto+).
        # @return [Symbol] Parameters parameter name.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def parameters_parameter_name # :nodoc:
          return :uuDto
        end

      end
    end
  end
end
