require 'timeout'
require 'securerandom'
require 'uu/os/ipc/ipc_error'
require 'uu/os/ipc/error_kind'

module UU
  module OS
    module Util

      # @private
      class ApplicationContainerService


        # Store reference to server context to be able to
        # provide various information.
        # @private
        def self.on_server_startup(ctx) # :nodoc:
          @@ctx = ctx
        end

        # Returns health info.
        # @private
        def on_health(ctx) # :nodoc:
          start = Thread.current[:start]
          entity = ctx.uu_uri
          result = {reqId: SecureRandom.uuid()}
          result[:terrUri] = entity
          result[:trace] = []
          invocation = {cont: get_container_code(ctx)}
          invocation[:nodeId] = get_node_id(ctx)
          invocation[:timestamp] = start
          result[:trace] << invocation
          invocation[:duration] = ((Time.now - start) * 1_000).to_i
          return result
        end

        # Returns server status.
        # @private
        def on_status(ctx) # :nodoc:
          if !@@ctx.test_or_development_mode?
            raise UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, nil, "Only available in development mode.")
          end
          status = {}
          if RUBY_ENGINE == 'jruby'
            status[:runtime] = "JRuby #{JRUBY_VERSION} (Ruby #{RUBY_VERSION}); Java #{java.lang.System.getProperty('java.version')} (#{java.lang.System.getProperty('java.vm.name')})"
          else
            status[:runtime] = "Ruby #{RUBY_VERSION}"
          end
          status[:uptime] = (Time.now - @@ctx.boot_time).to_i
          status[:libraries] = available_gems
          status[:configuration] = @@ctx.configuration
          status[:services] = @@ctx.registered_servlets
          return status
        end

        # Returns node ID.
        # @private
        def get_node_id(ctx) # :nodoc:
          node_id = ENV['UU_NODE_ID']
          if node_id.nil? && Object.const_defined?(:Java)
            node_id = Java::JavaLang::System.getProperty('uu.node.id')
          end
          if node_id.nil?
            node_id = ctx.request.headers[:host][0]
          end
          return node_id
        end

        def get_container_code(ctx)
          'APPLICATION'
        end

        # Returns list of available gems.
        # @private
        def available_gems # :nodoc:
          Gem::Specification.all = nil
          gems = Gem::Specification.map{|gem| "#{gem.name}:#{gem.version}"}
          Gem::Specification.reset
          gems
        end

      end

    end
  end
end